#include "circuit.hpp"

bool Circuit::check_if_supernodes_connected(SuperNode *s_node_1,
                                            SuperNode *s_node_2)
{
  bool ret = false;
  for (auto &node : *s_node_1) {
    auto it = find(s_node_2->begin(), s_node_2->end(), node);
    if (it != s_node_2->end()) {
      s_node_2->erase(it);
      ret = true;
    }
  }

  return ret;
}

double Circuit::calc_node_pair_conductance(Node *node1, Node *node2)
{
  double G = 0;
  if (node1 != node2) {
    for (auto &rstr : resistors)
      if ((rstr->get_node1() == node1 && rstr->get_node2() == node2) ||
          (rstr->get_node1() == node2 && rstr->get_node2() == node1))
        G -= rstr->get_conductance();
  } else {
    for (auto &rstr : resistors)
      if (rstr->get_node1() == node1 || rstr->get_node2() == node1)
        G += rstr->get_conductance();
  }

  return G;
}

Circuit::Circuit(Parser *parser) : nodes_num(parser->nodes.size())
{
  this->t_step = parser->t_step;
  this->resistors = parser->resistors;
  this->d_impdnc_devs = parser->d_impdnc_devs;
  this->nn_devs = parser->nn_devs;
  this->in_srcs = parser->in_srcs;
  this->nodes = parser->nodes;

  cnst_src_vec = ConstantSourceVector(nodes_num - 1, 0);

  for (int i = 0; i < nodes_num - 1; i++)
    cndct_mat.push_back(vector<double>(nodes_num - 1, 0));

  calc_non_linear_eq_components(vector<double>(nodes.size() - 1, 0));
  update_conductance_mat();
  update_constant_source_vec(0);
  calc_dynamic_parallel2Vsrc();
  // run calc_dynamic_eq_components on Solver (in order to be able to calc bias_point)
  // calc_dynamic_eq_components(vector<double>(nodes.size() - 1, 0));
}

void Circuit::calc_dynamic_parallel2Vsrc()
{
  Node *node1, *node2;
  Designator dsgntr;
  bool parallel2Vsrc;

  for (auto &d_impdnc_dev : d_impdnc_devs) {
    node1 = d_impdnc_dev->get_node1();
    node2 = d_impdnc_dev->get_node2();

    // check if there is a voltage source in the same nodes as the capacitor
    parallel2Vsrc = false;
    for (auto &in_src : in_srcs) {
      if (in_src->get_designator() == Designator::Voltage) {
        if (in_src->get_node1() == node1 &&
            in_src->get_node2() == node2) {
          if (dsgntr == Designator::Inductor)
            d_impdnc_dev->update_current(CIRCUIT_MAX_CURRENT);
          // there is already a voltage source there
          d_impdnc_dev->set_parallel2Vsrc(true);
          break;
        } else if (in_src->get_node2() == node1 &&
                   in_src->get_node1() == node2) {
          if (dsgntr == Designator::Inductor)
            d_impdnc_dev->update_current(-CIRCUIT_MAX_CURRENT);
          // there is already a voltage source there
          d_impdnc_dev->set_parallel2Vsrc(true);
          break;
        }
      }
    }
  }
}

void Circuit::set_unset_inductor_eq_shortcircuit()
{
  Node *node1, *node2;
  Designator dsgntr;

  for (auto &d_impdnc_dev : d_impdnc_devs) {
    node1 = d_impdnc_dev->get_node1();
    node2 = d_impdnc_dev->get_node2();
    dsgntr = d_impdnc_dev->get_designator();

    if (d_impdnc_dev->get_parallel2Vsrc())
      continue;

    if (dsgntr == Designator::Inductor) {
      InputSource *eq_source = d_impdnc_dev->get_eq_source();
      if (eq_source == NULL) {
        // Set equivalent source as a voltage source with V = 0
        VoltageSource *vs = new VoltageSource(node1, node2, 0);
        in_srcs.push_back(vs);
        d_impdnc_dev->set_eq_source(vs);
      } else if (eq_source->get_designator() == Designator::Voltage) {
        // if the equivalent shortcircuit source is set, find it and remove it
        auto it = find(in_srcs.begin(), in_srcs.end(), eq_source);
        if (it != in_srcs.end()) {
          d_impdnc_dev->set_eq_source(NULL);
          in_srcs.erase(it);
          node1->restore_is_connected_to_Vsrc();
          node2->restore_is_connected_to_Vsrc();
        } else {
          throw logic_error("SHOULD NOT BE HERE");
        }
      }
    }
  }
}

/******* Calculate Non Linear and Dynamic Devices' Equivalent Components ******/
void Circuit::calc_dynamic_eq_sources(vector<double> V)
{
  Designator dsgntr;
  Node *node1, *node2;
  double voltage, current;

  for (auto &d_impdnc_dev : d_impdnc_devs) {
    dsgntr = d_impdnc_dev->get_designator();
    node1 = d_impdnc_dev->get_node1();
    node2 = d_impdnc_dev->get_node2();

    if (dsgntr == Designator::Inductor || dsgntr == Designator::Capacitor)
      // DYNAMIC IMPEDANCE DEVICE //
      tie(voltage, current) = d_impdnc_dev->update_V_I(V, t_step);

    if (d_impdnc_dev->get_parallel2Vsrc())
      continue;

    if (dsgntr == Designator::Inductor) {
      // INDUCTOR //
      if (d_impdnc_dev->get_eq_source() == NULL) {
        // has not been initialized
        CurrentSource *cs = new CurrentSource(
          node1, node2, d_impdnc_dev->current);
        Resistor *rs = new Resistor(
          node1, node2, d_impdnc_dev->get_value() / t_step);
        resistors.push_back(rs);
        in_srcs.push_back(cs);
        d_impdnc_dev->set_eq_source(cs);
      } else {
        // has been initialized
        d_impdnc_dev->get_eq_source()->update_offset(d_impdnc_dev->current);
      }
    } else if (dsgntr == Designator::Capacitor) {
      // CAPACITOR //
      if (d_impdnc_dev->get_eq_source() == NULL) {
        // has not been initialized
        CurrentSource *cs = new CurrentSource(
          node1, node2, -voltage * d_impdnc_dev->get_value() / t_step);
        Resistor *rs = new Resistor(
          node1, node2, t_step / d_impdnc_dev->get_value());
        in_srcs.push_back(cs);
        resistors.push_back(rs);
        d_impdnc_dev->set_eq_source(cs);
      } else {
        // has been initialized
        d_impdnc_dev->get_eq_source()->update_offset(
          -voltage * d_impdnc_dev->get_value() / t_step);
      }
    }
  }
}

void Circuit::calc_non_linear_eq_components(vector<double> V)
{
  for (auto *nn_dev : nn_devs) {
    if (nn_dev->get_designator() == Designator::Diode) {
      auto prev_i_src = nn_dev->get_eq_i_source();
      auto prev_res = nn_dev->get_eq_resistor();
      nn_dev->update_eq_components(V);
      if (prev_i_src == NULL)
        in_srcs.push_back(nn_dev->get_eq_i_source());
      if (prev_res == NULL)
        resistors.push_back(nn_dev->get_eq_resistor());
    } else if (nn_dev->get_designator() == Designator::Transistor) {
      // nodes order -> collector base emmiter
      auto prev_i_src = nn_dev->get_eq_i_source();
      auto prev_diode = nn_dev->get_eq_diode();
      nn_dev->update_eq_components(V);
      if (prev_i_src == NULL)
        in_srcs.push_back(nn_dev->get_eq_i_source());
      if (prev_diode == NULL) {
        resistors.push_back(nn_dev->get_eq_diode()->get_eq_resistor());
        in_srcs.push_back(nn_dev->get_eq_diode()->get_eq_i_source());
        nn_devs.push_back(nn_dev->get_eq_diode());
      }
    }
  }
}


/********************** Circuit Matrix and Vector Update **********************/

ConductanceMatrix Circuit::update_conductance_mat()
{
  /**
   * Run anytime the components vector is being updated
   */

  Nodes available_nodes;

  for (int i = 0; i < cndct_mat.size(); i++)
    fill(cndct_mat[i].begin(), cndct_mat[i].end(), 0);

  create_supernodes(in_srcs);

  for (auto &node : nodes) {
    if (node->id == 0)
      reference = node;
    else if (!node->is_connected_to_Vsrc())
      update_with_simple_nodes(node);
    else
      available_nodes.push_back(node);
  }

  for (auto &in_src : in_srcs) {
    if (available_nodes.size()) {
      if (in_src->get_designator() == Designator::Voltage) {
        update_with_volt_source(in_src, available_nodes.back()->id);
        available_nodes.pop_back();
      }
    }
  }

  for (auto &s_node : s_nodes)
    if (available_nodes.size())
      if (update_with_supernode(s_node, available_nodes.back()->id))
        available_nodes.pop_back();

  return cndct_mat;
}

ConstantSourceVector Circuit::update_constant_source_vec(double t)
{
  Nodes available_nodes;

  create_supernodes(in_srcs);

  fill(cnst_src_vec.begin(), cnst_src_vec.end(), 0);

  for (auto &node : nodes)
    if (node->id == 0)
      reference = node;
    else if (!node->is_connected_to_Vsrc())
      update_const_with_simple_nodes(node, t);
    else
      available_nodes.push_back(node);

  for (auto &in_src : in_srcs) {
    if (available_nodes.size()) {
      if (in_src->get_designator() == Designator::Voltage) {
        update_const_with_volt_source(in_src, available_nodes.back()->id, t);
        available_nodes.pop_back();
      }
    }
  }

  for (auto &s_node : s_nodes)
    if (available_nodes.size())
      if (update_const_with_supernode(s_node, available_nodes.back()->id, t))
        available_nodes.pop_back();
  
  return cnst_src_vec;
}

/**************** Not Connected to Voltage Source Nodes Update ****************/

void Circuit::update_with_simple_nodes(Node *node)
{
  for (auto nnode : nodes)
    if (nnode->id != 0)
      cndct_mat[node->id - 1][nnode->id - 1] = calc_node_pair_conductance(node,
                                                                          nnode);
}

void Circuit::update_const_with_simple_nodes(Node *node, double t)
{
  for (auto &in_src : in_srcs) {
    if (in_src->get_designator() != Designator::Current)
      continue;
    if (in_src->get_node2() == node)
      cnst_src_vec[node->id - 1] += in_src->get_value(t);
    else if (in_src->get_node1() == node)
      cnst_src_vec[node->id - 1] -= in_src->get_value(t);
  }
}

/**************************** Voltage Source Update ***************************/

void Circuit::update_with_volt_source(InputSource * Vsrc, int row)
{
  Node *node1 = Vsrc->get_node1();
  Node *node2 = Vsrc->get_node2();

  // KVL for the voltage source
  for (auto &node : nodes) {
    int nid = node->id;

    if (nid == 0)
      continue;

    if (node == node1)
      cndct_mat[row - 1][nid - 1] = 1;
    else if (node == node2 && !node2->is_reference())
      cndct_mat[row - 1][nid - 1] = -1;
    else
      cndct_mat[row - 1][nid - 1] = 0;
  }
}

void Circuit::update_const_with_volt_source(InputSource * Vsrc, int row,
                                            double t)
{
  cnst_src_vec[row - 1] = Vsrc->get_value(t);
}

/****************************** Supernodes Create *****************************/

void Circuit::create_supernodes(InputSources aug_in_srcs)
{
  //make sure s_nodes is empty
  for (int i = 0; i < s_nodes.size(); i++) {
    s_nodes.erase(s_nodes.begin() + i);
  }

  // supernode for each voltage source
  for (auto &in_src : aug_in_srcs) {
    if (in_src->get_designator() == Designator::Voltage) {
      s_nodes.push_back(new SuperNode{in_src->get_node1(),
                                      in_src->get_node2()});
    }
  }

  // merge supernodes
  for (int i = 0; i < s_nodes.size(); i++) {
    for (int j = i + 1; j < s_nodes.size(); j++) {
      if (check_if_supernodes_connected(s_nodes[i], s_nodes[j])) {
        s_nodes[i]->insert(s_nodes[i]->end(),
                          s_nodes[j]->begin(), s_nodes[j]->end());
        s_nodes.erase(s_nodes.begin() + j);
        j--;
      }
    }
  }
}

/****************************** Supernodes Update *****************************/

bool Circuit::update_with_supernode(SuperNode *s_node, int row)
{
  fill(cndct_mat[row - 1].begin(), cndct_mat[row - 1].end(), 0);

  if (find(s_node->begin(), s_node->end(), reference) != s_node->end())
    return false;

  for (auto &node : nodes) {
    if (node == reference)
      continue;

    if (find(s_node->begin(), s_node->end(), node) != s_node->end()) {
      for (auto &rstr : resistors) {
        if ((find(s_node->begin(), s_node->end(), rstr->get_node1()) != s_node->end()) &&
            (find(s_node->begin(), s_node->end(), rstr->get_node2()) != s_node->end()))
          continue;

        if (rstr->get_node1() == node || rstr->get_node2() == node)
          cndct_mat[row - 1][node->id - 1] += rstr->get_conductance();
      }
    } else {
      for (auto &node_check : *s_node) {
        if (node_check->has_neighbor(node))
          cndct_mat[row - 1][node->id - 1] += calc_node_pair_conductance(node_check,
                                                                         node);
      }
    }
  }

  return true;
}

bool Circuit::update_const_with_supernode(SuperNode *s_node, int row, double t)
{
  if (find(s_node->begin(), s_node->end(), reference) != s_node->end())
    return false;

  cnst_src_vec[row - 1] = 0;
  for (auto &in_src : in_srcs) {
    if (in_src->get_designator() != Designator::Current)
      continue;

    if (find(s_node->begin(), s_node->end(), in_src->get_node2()) != s_node->end())
      cnst_src_vec[row - 1] += in_src->get_value(t);
    if (find(s_node->begin(), s_node->end(), in_src->get_node1()) != s_node->end())
      cnst_src_vec[row - 1] -= in_src->get_value(t);
  }

  return true;
}


/*********************************** GET-ers **********************************/

ConstantSourceVector Circuit::get_const_source_vec()
{
  return cnst_src_vec;
}

ConductanceMatrix Circuit::get_conductance_mat()
{
  return cndct_mat;
}

InputSources Circuit::get_input_sources()
{
  return in_srcs;
}

SuperNodes Circuit::get_supernodes()
{
  return s_nodes;
}