#ifndef COMPONENTS_H
#define COMPONENTS_H

#include "universal.hpp"
#include "node.hpp"

#define DIODE_Is 1e-11
#define DIODE_VT 25e-3
#define TRANSISTOR_B 100

enum class Designator
{
  Voltage,
  Current,
  Capacitor,
  Inductor,
  Resistor,
  Diode,
  Transistor
};

class Component
{
public:
  Component(Designator dsgntr, Node *nd1, Node *nd2, Node *nd3);

  Component(Designator dsgntr, Node *nd1, Node *nd2);

  Node* get_node1();
  Node* get_node2();

  Designator get_designator();
  double calc_voltage(vector<double> V);
  double get_current();
  virtual pair<double, double> update_V_I(vector<double> V);
  virtual pair<double, double> update_V_I(vector<double> V, double t_step);

  double current = 0;
  double voltage = 0;
protected:
  Designator designator;
  Node *node1, *node2, *node3;
};

class InputSource : public Component
{
public:
  InputSource(Designator dsgntr, Node *nd1, Node *nd2, double offset,
              double Amplitude, double freq);

  double get_value(double t) const;
  double get_offset() const;
  double get_frequency() const;
  double get_amplitude() const;
  void update_offset(double offset);

// private:
  double A, off, f;
};

class VoltageSource : public InputSource
{
public:
  VoltageSource(Node *nd1, Node *nd2, double offset, double Amplitude = 0,
                double freq = 0);
};

class CurrentSource : public InputSource
{
public:
  CurrentSource(Node *nd1, Node *nd2, double offset, double Amplitude = 0,
                double freq = 0);
};

class ImpedanceDevice : public Component
{
public:
  ImpedanceDevice(Designator dsgntr, Node *nd1, Node *nd2, double vl);

  double get_value();
  void update_value(double value);
  double get_voltage();

// protected:
  double value;
};

class Resistor : public ImpedanceDevice
{
public:
  Resistor(Node *nd1, Node *nd2, double vl);
  double get_conductance();

  pair<double, double> update_V_I(vector<double> V);
};

class DynamicImpedanceDevice : public ImpedanceDevice
{
public:
  DynamicImpedanceDevice(Designator dsgntr, Node *nd1, Node *nd2, double vl);

  virtual void set_eq_source(InputSource *eq_source);
  virtual InputSource *get_eq_source();
  void update_current(double current, double prev_current = 0);
  void set_parallel2Vsrc(bool parallel2Vsrc);
  bool get_parallel2Vsrc();

  double prev_voltage = 0;
  double prev_current = 0;

protected:
  bool parallel2Vsrc = false;
  InputSource *eq_source = NULL;
};

class Capacitor : public DynamicImpedanceDevice
{
public:
  Capacitor(Node *nd1, Node *nd2, double vl);

  void set_eq_source(InputSource *eq_source);
  InputSource *get_eq_source();
  pair<double, double> update_V_I(vector<double> V, double t_step);
};

class Inductor : public DynamicImpedanceDevice
{
public:
  Inductor(Node *nd1, Node *nd2, double vl);

  void set_eq_source(InputSource *eq_source);
  InputSource *get_eq_source();
  pair<double, double> update_V_I(vector<double> V, double t_step);
};


class NonLinearDevice : public Component
{
public:
  NonLinearDevice(Designator dsgntr, Node *nd1, Node *nd2, Node *nd3,
                  string vl);
  NonLinearDevice(Designator dsgntr, Node *nd1, Node *nd2, string vl);

  string get_type() const;

  void set_eq_components(Resistor *eq_resistor, CurrentSource *eq_i_source);
  void set_eq_components(NonLinearDevice *eq_diode, CurrentSource *eq_i_source);
  virtual void update_eq_components(vector<double> V);

  NonLinearDevice *get_eq_diode();
  Resistor *get_eq_resistor();
  CurrentSource *get_eq_i_source();

// protected:
  string value;

protected:
  NonLinearDevice *eq_diode = NULL;
  Resistor *eq_resistor = NULL;
  CurrentSource *eq_i_source = NULL;
};

class Transistor : public NonLinearDevice
{
public:
  Transistor(Node *nd1, Node *nd2, Node *nd3, string vl);
  double calc_voltage(vector<double> V);
  void update_eq_components(vector<double> V);
};

class Diode : public NonLinearDevice
{
public:
  Diode(Node *nd1, Node *nd2);
  void update_eq_components(vector<double> V);
};

#endif