#include "parser.hpp"

Parser::Parser (string filename)
{
  fstream newfile;

  newfile.open(filename,ios::in);
  if (newfile.is_open()) {
    string tp;
    while(getline(newfile, tp)) {
      if (tp[0] == 'R')
        resistors.push_back(parse_resistor_line(tp));
      if (tp[0] == 'C' || tp[0] == 'L')
        d_impdnc_devs.push_back(parse_dynamic_impedance_line(tp));
      if (tp[0] == 'V' || tp[0] == 'I') {
        in_srcs.push_back(parse_source_line(tp));
      }
      if (tp[0] == 'Q' || tp[0] == 'D')
        nn_devs.push_back(parse_non_linear_line(tp));
      if (tp.substr(0, 5) == ".tran")
        tie(t_stop, t_step) = parse_time(tp);
      if (tp.substr(0, 4) == ".end")
        break;
    }
    newfile.close();
  }

  for (auto &nd : nodes)
    nd->extend_neighbors(nodes.size());
}

tuple<char, Node*, Node*, double> Parser::parse_impedance_line (string line)
{
  double base_value, multiplier;
  char dev_type;

  Node *node1, *node2;

  size_t pos = 0;
  vector<string> tokens;
  while ((pos = line.find(" ")) != string::npos) {
      tokens.push_back(line.substr(0, pos));
      line.erase(0, pos + 1);
  }
  tokens.push_back(line);

  node1 = get_node(stoi("0" + tokens[1].substr(1)));
  node2 = get_node(stoi("0" + tokens[2].substr(1)));

  dev_type = tokens[0][0];

  base_value = stod(tokens[3]);
  multiplier = calc_mult(tokens[3].substr(nb_digits(base_value))[0]);

  return make_tuple(dev_type, node1, node2, base_value * pow(10, multiplier));
}

Resistor * Parser::parse_resistor_line (string line)
{
  Node *node1, *node2;
  double value;
  char dev_type;

  tie(dev_type, node1, node2, value) = parse_impedance_line(line);

  node1->update_neighbors(node2);
  node2->update_neighbors(node1);

  return new Resistor(node1, node2, value);
}

DynamicImpedanceDevice * Parser::parse_dynamic_impedance_line (string line)
{
  DynamicImpedanceDevice *d_impdnc_dev;
  Node *node1, *node2;
  double value;
  char dev_type;

  tie(dev_type, node1, node2, value) = parse_impedance_line(line);

  switch (dev_type) {
    case 'C':
      d_impdnc_dev = new Capacitor(node1, node2, value);
      break;
    case 'L':
      d_impdnc_dev = new Inductor(node1, node2, value);
      break;
  }

  return d_impdnc_dev;
}

InputSource * Parser::parse_source_line (string line)
{
  InputSource *src;

  double base_value, multiplier,
          base_off, mult_off,
          base_A, mult_A,
          base_f, mult_f;
  char dev_type;
  bool is_AC_source;

  Node *node1, *node2;

  size_t pos = 0;
  vector<string> tokens;
  while ((pos = line.find(" ")) != string::npos) {
      tokens.push_back(line.substr(0, pos));
      line.erase(0, pos + 1);
  }
  tokens.push_back(line);
  
  node1 = get_node(stoi("0" + tokens[1].substr(1)));
  node2 = get_node(stoi("0" + tokens[2].substr(1)));

  dev_type = tokens[0][0];

  is_AC_source = tokens[3].substr(0, 4) == "SINE";

  if (! is_AC_source) {
    base_value = stod(tokens[3]);
    multiplier = calc_mult(tokens[3].substr(nb_digits(base_value))[0]);
  } else {
    base_off = stod(tokens[3].substr(5));
    mult_off = calc_mult(tokens[3].substr(nb_digits(base_off))[0]);

    base_A = stod(tokens[4]);
    mult_A = calc_mult(tokens[4].substr(nb_digits(base_A))[0]);

    base_f = stod(tokens[5]);
    mult_f = calc_mult(tokens[5].substr(nb_digits(base_f))[0]);
  }

  switch (dev_type) {
    case 'V':
      if (! is_AC_source) {
        src = new VoltageSource(node1, node2, base_value*pow(10, multiplier));
        break;
      }
      src = new VoltageSource(node1, node2, base_off * pow(10, mult_off),
                              base_A*pow(10, mult_A), base_f*pow(10, mult_f));
     
      break;
    case 'I':
      if (! is_AC_source) {
        src = new CurrentSource(node1, node2, base_value*pow(10, multiplier));
        break;
      }
      src = new CurrentSource(node1, node2, base_off*pow(10, mult_off),
                              base_A*pow(10, mult_A), base_f*pow(10, mult_f));
      break;
  }
  return src;
}

NonLinearDevice * Parser::parse_non_linear_line (string line)
{
  NonLinearDevice *dev;

  char dev_type;

  Node *node1, *node2, *node3;

  size_t pos = 0;
  vector<string> tokens;
  while ((pos = line.find(" ")) != string::npos) {
      tokens.push_back(line.substr(0, pos));
      line.erase(0, pos + 1);
  }
  tokens.push_back(line);

  node1 = get_node(stoi("0" + tokens[1].substr(1)));
  node2 = get_node(stoi("0" + tokens[2].substr(1)));

  dev_type = tokens[0][0];

  switch (dev_type) {
    case 'D':
      dev = new Diode(node1, node2);
      break;
    case 'Q':
      node3 = get_node(stoi("0" + tokens[3].substr(1)));
      dev = new Transistor(node1, node2, node3, tokens[4]);
      break;
  }
  return dev;
}

pair<double, double> Parser::parse_time (string line)
{
  size_t pos = 0;
  vector<string> tokens;

  while ((pos = line.find(" ")) != string::npos) {
       tokens.push_back(line.substr(0, pos));
       line.erase(0, pos + 1);
  }
  tokens.push_back(line);

  double base_t_stop = stod(tokens[2]);
  double mult_t_stop = calc_mult(tokens[2].substr(nb_digits(base_t_stop))[0]);

  double t_stop = base_t_stop*pow(10, mult_t_stop);

  double base_t_step = stod(tokens[4]);
  double mult_t_step = calc_mult(tokens[4].substr(nb_digits(base_t_step))[0]);

  double t_step = base_t_step*pow(10, mult_t_step);

  return make_pair(t_stop, t_step);
}

int Parser::nb_digits(double value) {
  string s = to_string(value);
  s.erase(s.find_last_not_of('0') + 1, string::npos);
  if (!s.empty() && !isdigit(s.back()))
    s.pop_back();

  return s.length();
}

Node * Parser::get_node(int node_id)
{
  auto it = find_if(nodes.begin(), nodes.end(),
                    [node_id](const Node * c)
                    {return c->id == node_id; });

  Node *node;
  if (it == nodes.end()) {
    node = new Node(node_id, false);
    nodes.push_back(node);
  } else {
    node = *it;
  }

  return node;
}

int Parser::calc_mult(char multipl)
{
  switch (multipl)
  {
    case 'p':
      return -12;
    case 'n':
      return -9;
    case 'u':
      return -6;
    case 'm':
      return -3;
    case 'k':
      return 3;
    case 'M':
      return 6;
    case 'G':
      return 9;
    default: // '\0'
      return 0;
  }
}

// int main()
// {
//   Parser parser("parser.txt");

//   cout << parser.nodes << endl;
//   cout << endl;
//   cout << parser.impdnc_devs << endl;
//   cout << endl;
//   cout << parser.in_srcs << endl;
//   cout << endl;
//   cout << parser.nn_devs << endl;
// }