#include "universal.hpp"

#include "node.hpp"
#include "components.hpp"
#include "circuit.hpp"
#include "parser.hpp"

using namespace std;
// using namespace Eigen;

class Solver
{
private:
  const int nodes_num;
  double threshold, bias_calc_lr, conv_patience;
  vector<Eigen::VectorXd> solution;
  Circuit *circuit;
  Eigen::MatrixXd G;

  template <typename T>
  void vec_merge(vector<T> *v1, vector<T> *v2)
  {
    v1->insert(v1->end(), v2->begin(), v2->end());
  }

public:
  double t_stop, t_step;

  Solver(Circuit *circuit, double t_stop, double t_step)
    : nodes_num(circuit->nodes_num)
  {
    this->circuit = circuit;
    this->t_stop = t_stop;
    this->t_step = t_step;
    this->threshold = 1e-3;
    this->bias_calc_lr = 1e-2;
    this->conv_patience = 100;
  }

  void simulate() {
    /** Simulate Circuit for Each Timestep
     * 
     * @see simulation_step
     */

    vector<double> voltages;
    cout << "In Simulation Time: 0" << endl;
    int cnt = 0;
    for (double t = 0; t <= t_stop; t += t_step) {
      if (cnt ++ == 100) {
        cout << "100 steps done --> simulation time: " << t << endl;
        cnt = 0;
      }

      solution.push_back(simulation_step(t));
      
      // update equivalent linear components in circuit based on current voltage
      voltages = eigen2std_vec(solution.back());

      circuit->calc_non_linear_eq_components(voltages);

      // this will integrate on the current of inductors' equal current sources
      circuit->calc_dynamic_eq_sources(voltages);

      if (t == 0) {
        for (auto &d_impdnc_dev : circuit->d_impdnc_devs)
          if (d_impdnc_dev->get_designator() == Designator::Inductor)
            iterative_calc_inductors_current(solution.back(), d_impdnc_dev);
      }
    }
  }

  void iterative_calc_inductors_current(
    Eigen::VectorXd bias_point_V, DynamicImpedanceDevice *indctr)
  {
    /** Run iterative method to calculate node voltages
     * 
     *************************** General Idea **********************************
     * last random guess --> solve --> new guess -calc diff-> until close to 0
     ***************************************************************************
     * 
     * @return -- Eigen::VectorXd with current step's node voltages
     * @param time -- Current time to simulate
     * @see simulation_step
     * @see solve_linear_system
     */

    if (indctr->get_eq_source() == NULL)
      return;

    cout << "Iteratively calculate inductors' initial currents" << endl;

    Eigen::VectorXd cur_sol;
    double current, prev_error, error, multiplier, cnt;
    do {
      current = 10;
      prev_error = 0;
      error = 0;
      multiplier = 1;
      cnt = 0;
      indctr->current = current;
      indctr->get_eq_source()->update_offset(current);

      circuit->update_conductance_mat();

      do {
        circuit->update_constant_source_vec(0);
        cur_sol = solve_linear_system();

        error = (bias_point_V - cur_sol).norm();

        if (error > prev_error)
          multiplier *= -1;

        if (error == prev_error)
          cnt++;
        else
          cnt = 0;

        if (cnt >= conv_patience)
          current = CIRCUIT_MAX_CURRENT *
                    signbit(indctr->calc_voltage(eigen2std_vec(bias_point_V)));
        else
          current = current - error * multiplier * bias_calc_lr;

        indctr->current = current;
        indctr->get_eq_source()->update_offset(current);
        prev_error = error;

        if (cnt >= conv_patience)
          break;

      } while ((cur_sol - bias_point_V).norm() > threshold);
      cout << "For lr = " << bias_calc_lr << " current is " << current << endl;
      bias_calc_lr /= 10;
    } while (isnan(current) || isinf(current));
    cout << "Iteratively calculate inductors' initial currents DONE" << endl;
  }

  Eigen::VectorXd simulation_step(double time)
  {
    /** Simulate the circuit for the current timestep.
     * 
     * if t = 0 calculate bias point
     * * inductors  = shortcircuit = voltage source wih zero voltage
     * * capacitors = opencircuit
     * 
     * @return -- Eigen::VectorXd with current step's node voltages
     * @param time -- Current time to simulate
     * @see iterative_solve_circuit
     * @see simulate
     */
    if (time == 0)
      circuit->set_unset_inductor_eq_shortcircuit();

    // because different time and different resistors
    // from non linear and dynamic components
    circuit->update_constant_source_vec(time);

    // because different resistors from non linear and dynamic components
    circuit->update_conductance_mat();

    Eigen::VectorXd step_sol = iterative_solve_circuit(time);

    if (time == 0)
      circuit->set_unset_inductor_eq_shortcircuit();

    return step_sol;
  }


  Eigen::VectorXd iterative_solve_circuit(double time)
  {
    /** Run Newton Raphson method to calculate node voltages
     * 
     *************************** General Idea **********************************
     * last random guess --> solve --> new guess -calc diff-> until close to 0
     ***************************************************************************
     * 
     * @return -- Eigen::VectorXd with current step's node voltages
     * @param time -- Current time to simulate
     * @see simulation_step
     * @see solve_linear_system
     */

    Eigen::VectorXd prev_sol, cur_sol;
    if (solution.size())
      cur_sol = solution.back();
    else
      cur_sol = Eigen::VectorXd::Zero(nodes_num - 1);

    vector<double> voltages;
    int i = 0;
    do {
      prev_sol = cur_sol;
      voltages = eigen2std_vec(prev_sol);

      /* Update equivalent linear components in circuit based on previous
       * voltage vector as an initial guess for the proper voltage vector. */
      circuit->calc_non_linear_eq_components(voltages);

      /* Each update must be follower by an update of the constant source vector
       * and the conductance matrix because different time and different
       * resistors from non linear and dynamic components */
      circuit->update_constant_source_vec(time);
      circuit->update_conductance_mat();

      /* Then, we solve the circuit for this voltage vector guess */
      cur_sol = solve_linear_system();

      // if (i++ >= 100)
      //   break;
    } while ((cur_sol - prev_sol).norm() > threshold);
    // } while ((cur_sol - prev_sol).norm() > cur_sol.norm()*0.001 + 1e-4);
    return cur_sol;
  }

  Eigen::VectorXd solve_linear_system()
  {
    G = std2eigen_mat(circuit->get_conductance_mat());
    Eigen::VectorXd cs = std2eigen_vec(circuit->get_const_source_vec());
    Eigen::VectorXd step_sol = G.colPivHouseholderQr().solve(cs);
    // cout << G << endl;
    // cout << cs << endl;
    // cout << step_sol << endl << endl;
    return step_sol;
  }

  vector<Eigen::VectorXd> get_solution() {
    return solution;
  }

  Eigen::MatrixXd get_conductance_matrix() {
    return G;
  }

  Eigen::MatrixXd std2eigen_mat(vector<vector<double>> std_mat)
  {
    vector<double> temp;

    for (auto &mat_row : std_mat)
      vec_merge(&temp, &mat_row);

    Eigen::MatrixXd eigen_mat;
    eigen_mat = Eigen::Map<Eigen::MatrixXd>(temp.data(), std_mat.size(),
                                            std_mat[0].size());
    eigen_mat.transposeInPlace();

    return eigen_mat;
  }

  Eigen::VectorXd std2eigen_vec(vector<double> std_vec)
  {
    Eigen::VectorXd eigen_vec = Eigen::VectorXd::Ones(std_vec.size());
    for (int i = 0; i < std_vec.size(); i++)
      eigen_vec(i) = std_vec[i];
    // Eigen::Map<Eigen::VectorXd> eigen_vec(std_vec.data(), std_vec.size());
    return eigen_vec;
  }

  vector<double> eigen2std_vec(Eigen::VectorXd eigen_vec)
  {
    vector<double> std_vec;
    std_vec.resize(eigen_vec.size());
    Eigen::VectorXd::Map(&std_vec[0], eigen_vec.size()) = eigen_vec;

    return std_vec;
  }
};